/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        defedit.h ( Live Once Library, C++ )
 *
 * COMMENTS:
 */

#include <libtcod.hpp>

#include "defedit.h"
#include "chooser.h"
#include "gfxengine.h"


// Prototypes stolen from Main
extern int popupText(const char *buf, int delayms = 0);
extern int popupText(BUF buf, int delayms = 0);
extern BUF popupQuestion(const char *buf);
extern BUF popupQuestion(BUF buf);
extern void redrawWorld();
extern void editLook(const char *thing, BUF &name, u8 &symbol, ATTR_NAMES attr_name);

extern CHOOSER		*glbChooser;
extern bool		 glbChooserActive;

void
editRangeAttack(MOB *mob)
{
    MOB_DEF		*def = GAMEDEF::mobdef(mob->getDefinition());

    const int lastchoice = 4;
    int	choice = 4;

    const char *choices[] =
    {
	"Edit Look",
	"Change Range",
	"Change Recharge",
	"Change Attack",
	"Done",
	0
    };

    int		level = mob->getLevel();

    // Get an attackdef.
    if (def->range_attack == ATTACK_NONE)
    {
	def->range_attack = GAMEDEF::createNewAttackDef();
    }
    ATTACK_DEF *attackdef = GAMEDEF::attackdef(def->range_attack);

    while (1)
    {
	mob->setupTextVariables();

	BUF	query = text_lookup("query", "mob_editrange");

	if (!def->range_range)
	{
	    query.strcat("No Ranged Attack");
	}
	else
	{
	    BUF		stats;
	    stats.sprintf("Range: %d; Recharge: %d\n",
			def->range_range, def->range_recharge);
	    query.strcat(stats);

	    stats = MOB::buildAttackDescr((ATTACK_NAMES)def->range_attack);
	    if (stats.isstring())
	    {
		query.strcat("\n");
		query.strcat(stats);
	    }
	}


	choice = glbEngine->askQuestionList(query, choices, choice, true);

	if (choice < 0)
	    break;

	def->set_range = true;

	if (choice == lastchoice)
	    break;

	switch (choice)
	{
	    case 0:
	    {
		BUF		newname;
		u8		symbol = def->range_symbol;
		ATTR_NAMES	look = (ATTR_NAMES) def->range_attr;
		if (look < NUM_ATTRS)
		{
		    // Unique!
		    ATTR_NAMES	nlook = GAMEDEF::createNewAttrDef();
		    ATTR_DEF	*nattr = GAMEDEF::attrdef(nlook);
		    *nattr = *GAMEDEF::attrdef(look);
		    def->range_attr = nlook;
		    look = nlook;
		}

		newname.strcpy(attackdef->noun);
		editLook("range attack", newname, symbol, look);
		if (newname.strcmp(attackdef->noun))
		    attackdef->noun = glb_harden(newname);
		def->range_symbol = symbol;
		break;
	    }
	    case 1:
	    {
		def->range_range = defedit_chooseRange(true);
		if (def->range_range)
		    def->range_valid = true;
		else
		    def->range_valid = false;
		break;
	    }	
	    case 2:
	    {
		def->range_recharge = defedit_chooseRecharge(true);
		break;
	    }
	    case 3:
	    {
		defedit_editAttack(mob, 0, (ATTACK_NAMES) def->range_attack, true);
		break;
	    }
	}
    }

}

void
editSpell(BUF prologue, SPELL_NAMES spell)
{
    SPELL_DEF	*def = GAMEDEF::spelldef(spell);

    const int lastchoice = 6;
    int	choice = 6;

    const char *choices[] =
    {
	"Edit Look",
	"Change Verb",
	"Change Effect",
	"Change Mana Use",
	"Change Area Effect",
	"Toggle Directed",
	"Done",
	0
    };

    int		level = GAMEDEF::itemdef(def->item)->depth;

    while (1)
    {
	BUF		query;
	query = prologue;

	if (query.isstring())
	    query.strcat("\n\n");

	query.strcat("What should happen when casting ");
	query.strcat(def->name);
	query.strcat("?\n");
	BUF stats = MOB::buildSpellDescr(spell);
	if (stats.isstring())
	{
	    query.strcat("\n");
	    query.strcat(stats);
	}

	choice = glbEngine->askQuestionList(query, choices, choice, true);

	if (choice < 0)
	    break;
	if (choice == lastchoice)
	    break;

	switch (choice)
	{
	    case 0:
	    {
		BUF		newname;
		u8		symbol = def->symbol;
		ATTR_NAMES	look = (ATTR_NAMES) def->attr;
		if (look < NUM_ATTRS)
		{
		    // Unique!
		    ATTR_NAMES	nlook = GAMEDEF::createNewAttrDef();
		    ATTR_DEF	*nattr = GAMEDEF::attrdef(nlook);
		    *nattr = *GAMEDEF::attrdef(look);
		    def->attr = nlook;
		    look = nlook;
		}

		newname.strcpy(def->name);
		editLook("spell", newname, symbol, look);
		if (newname.strcmp(def->name))
		    def->name = glb_harden(newname);
		def->symbol = symbol;
		break;
	    }
	    case 1:
	    {
		BUF	query, response;
		query.sprintf("What verb should be used when creatures are hit with this spell?");
		response = glbEngine->askQuestionText(query, true);
		if (response.isstring())
		    def->verb = glb_harden(response);
		break;
	    }	
	    case 2:
	    {
		// CHange effect.
		def->effect = defedit_editEffect((EFFECT_NAMES)def->effect, level, true);
		break;
	    }
	    case 3:
	    {
		float scale = defedit_chooseAmount("How much mana should this spell use?", true);
		def->mana = MOB::expectedMPforLevel(level) * (scale/3);
		if (def->mana < 1)
		    def->mana = 1;
		break;
	    }
	    case 4:
	    {
		def->radius = defedit_chooseAreaEffect(true);
		break;
	    }
	    case 5:
	    {
		def->needsdir = !def->needsdir;
		break;
	    }
	}
    }
}

void
editElements(bool isclient)
{
    int		choice;

    PTRLIST<BUF>	choices;
    PTRLIST<ELEMENT_NAMES>	elems;
    BUF	tmp;

    while (1)
    {
	choices.clear();
	elems.clear();
	for (int e = 0; e < GAMEDEF::getNumElement(); e++)
	{
	    if (e == ELEMENT_NONE)
		continue;

	    tmp.reference(GAMEDEF::elementdef(e)->name);
	    choices.append(tmp);
	    elems.append((ELEMENT_NAMES) e);
	}

	tmp.reference("Add New Element");
	choices.append(tmp);

	tmp.reference("Done");
	choices.append(tmp);

	choice = glbEngine->askQuestionListCanned("query", "edit_element", choices, choices.entries()-1, isclient);
	if (choice < 0)
	    return;

	if (choice == choices.entries()-1)
	{
	    // Done
	    return;
	}

	if (choice == choices.entries()-2)
	{
	    // Add a new element
	    int		nelem = GAMEDEF::createNewElementDef();
	    GAMEDEF::elementdef(nelem)->name = "phlogiston";
	}
	else
	{
	    // Rename!
	    BUF		question;

	    question.sprintf("What should be the new name for %s?",
				GAMEDEF::elementdef(elems(choice))->name);
	    BUF	response = glbEngine->askQuestionText(question, isclient);

	    if (response.isstring())
	    {
		GAMEDEF::elementdef(elems(choice))->name = glb_harden(response);
	    }
	}
    }
}

void
editAttack(BUF prologue, MOB *mob, ITEM *item, ATTACK_NAMES attack)
{
    ATTACK_DEF	*def = GAMEDEF::attackdef(attack);

    const int		lastchoice = 7;
    int			choice = lastchoice;
    ENGINE::Question		query;

    const char *choices[] =
    {
	"Change Noun",
	"Change Verb",
	"Damage",
	"Distribution",
	"Bonus to Hit",
	"Element",
	"Special Effect",
	"Done",
	0
    };

    query.setChoices(choices);
    query.type = ENGINE::QUERY_CHOICE;

    while (!TCODConsole::isWindowClosed())
    {
	query.query = prologue;
	query.query.strcat( MOB::buildAttackDescr(attack) );

	query.defaultchoice = choice;
	ENGINE::Response	response = defedit_askQuestion(query);

	if (response.type != ENGINE::QUERY_CHOICE)
	{
	    break;
	}

	// Edited something...
	def->set_special = true;

	choice = response.choice;

	if (choice == lastchoice)
	    break;

	switch (choice)
	{
	    case 0:
	    {
		BUF	query, response;
		query.sprintf("What should %s be called?", def->noun);
		response = glbEngine->askQuestionText(query, true);
		if (response.isstring())
		    def->noun = glb_harden(response);
		break;
	    }
	    case 1:
	    {
		BUF	query, response;
		query.sprintf("Instead of \"You %s with %s\", what verb do you wish?", def->verb, def->noun);
		response = glbEngine->askQuestionText(query, true);
		if (response.isstring())
		    def->verb = glb_harden(response);
		break;
	    }	
	    case 2:
		defedit_editAttackDamage(mob, item, attack, true);
		break;
	    case 3:
		defedit_editAttackDistribution(mob, item, attack, true);
		break;
	    case 4:
		defedit_editAttackBonusToHit(mob, item, attack, true);
		break;
	    case 5:
	    {
		ELEMENT_NAMES	elem = defedit_chooseElement((ELEMENT_NAMES)def->element, true);

		if (elem != ELEMENT_NONE)
		    def->element = elem;
		break;
	    }
	    case 6:
	    {
		BUF		prequel;
		prequel.reference("What extra thing happens with a successful attack?");
		int			level = 1;
		if (mob) level = mob->getLevel();
		if (item) level = item->getDepth();
		def->effect = defedit_editEffect(prequel,
				(EFFECT_NAMES)def->effect,
				level,
				true);
		break;
	    }
	}
    }
}

ENGINE::Response
defedit_askQuestion(const ENGINE::Question &query)
{
    ENGINE::Response	response;

    switch (query.type)
    {
	case ENGINE::QUERY_TEXT:
	{
	    BUF		text = query.query;
	    text.stripWS();
	    response.type = query.type;
	    response.answer = popupQuestion(text);
	    break;
	}
	case ENGINE::QUERY_CHOICE:
	{
	    glbChooser->clear();

	    glbChooser->setTextAttr(ATTR_NORMAL);
	    BUF		wrapped;

	    wrapped = query.query.wordwrap(SCR_WIDTH-15);
	    wrapped.stripWS();
	    wrapped.append('\n');
	    wrapped.append('\n');

	    glbChooser->setPrequel(wrapped);
	    glbChooser->setPrequelAttr(ATTR_WHITE);
	    for (int i = 0; i < query.choices.entries(); i++)
		glbChooser->appendChoice(query.choices(i));
	    glbChooser->setChoice(query.defaultchoice);

	    response.choice = -1;

	    // Run the chooser...
	    glbChooserActive = true;
	    while (!TCODConsole::isWindowClosed())
	    {
		redrawWorld();
		int key = gfx_getKey(false);

		glbChooser->processKey(key);

		if (key)
		{
		    // User selects this?
		    if (key == '\n' || key == ' ')
		    {
			response.choice = glbChooser->getChoice();
			break;
		    }
		    else if (key == '\x1b')
		    {
			// Choose nothing.
			break;
		    }
		}
	    }
	    glbChooserActive = false;
	    if (response.choice >= 0)
		response.type = ENGINE::QUERY_CHOICE;

	    break;
	}
	case ENGINE::QUERY_ATTACK:
	{
	    editAttack(query.query, query.mob, query.item, (ATTACK_NAMES)query.payload);
	    response.type = ENGINE::QUERY_ATTACK;
	    break;
	}
	case ENGINE::QUERY_SPELL:
	{
	    editSpell(query.query, (SPELL_NAMES)query.payload);
	    response.type = ENGINE::QUERY_SPELL;
	    break;
	}
	case ENGINE::QUERY_RANGEATTACK:
	{
	    editRangeAttack(query.mob);
	    response.type = ENGINE::QUERY_RANGEATTACK;
	    break;
	}
	case ENGINE::QUERY_INVALID:
	    // Leave default
	    break;
    }

    return response;
}

void
defedit_editMobHP(MOB *mob, bool isclient)
{
    const char *options[] =
    {
	"Windshield Kill",
	"Easy Kill",
	"Normal Health",
	"Extra Health",
	"Double Health",
	"Slugfest",
	0
    };

    mob->setupTextVariables();
    int	hpchoice = glbEngine->askQuestionListCanned("query", "mob_hp", options, 2, isclient);

    if (hpchoice >= 0)
    {
	mob->edefn().set_hp = true;
	int		nhp = mob->expectedHPforLevel(mob->getLevel());
	switch (hpchoice)
	{
	    case 0:
		nhp = rand_range(nhp/10, nhp/5);
		break;
	    case 1:
		nhp = rand_range(nhp/4, 3*nhp/4);
		break;
	    case 2:
		nhp = rand_range(3*nhp/4, 5*nhp/4);
		break;
	    case 3:
		nhp = rand_range(nhp, 2*nhp);
		break;
	    case 4:
		nhp = rand_range(2*nhp, 3*nhp);
		break;
	    case 5:
		nhp = rand_range(3*nhp, 5*nhp);
		break;
	}
	mob->edefn().max_hp = nhp;
	if (mob->pos().map())
	{
	    // Boost all of our type to max hp
	    MOBLIST		mobs;
	    mob->pos().map()->getAllMobsOfType(mobs, mob->getDefinition());
	    for (int i = 0; i < mobs.entries(); i++)
	    {
		mobs(i)->gainHP(nhp);
	    }
	}
    }
}

void
defedit_editMobAI(MOB *mob, bool isclient)
{
    mob->setupTextVariables();

    int		choice;
    const char *choices[] =
    {
	"Wander... Then CHARGE!",
	"Hang out at home... Then CHARGE!",
	"Charge! (but avoid diagonals)",
	"Track you down",
	"Try to out-flank",
	"Shoot from afar",
	"Run away",
	"Look for friends first",
	"Stick to walls",
	"Peaceful",
	0
    };

    choice = glbEngine->askQuestionListCanned("query", "mob_ai", choices, 0, isclient);
    if (choice < 0)
	return;

    AI_NAMES		newai = AI_NONE;
    switch (choice)
    {
	case 0:
	    newai = AI_CHARGE;
	    break;
	case 1:
	    newai = AI_HOME;
	    break;
	case 2:
	    newai = AI_ORTHO;
	    break;
	case 3:
	    newai = AI_PATHTO;
	    break;
	case 4:
	    newai = AI_FLANK;
	    break;
	case 5:
	    newai = AI_RANGECOWARD;
	    break;
	case 6:
	    newai = AI_COWARD;
	    break;
	case 7:
	    newai = AI_RAT;
	    break;
	case 8:
	    newai = AI_MOUSE;
	    break;
	case 9:
	    newai = AI_PEACEFUL;
	    break;
    }

    if (newai != AI_NONE)
    {
	mob->edefn().set_ai = true;
	mob->edefn().ai = newai;
    }
}

void
defedit_editMobRarity(MOB *mob, bool isclient)
{
    mob->setupTextVariables();

    int		choice;
    const char *choices[] =
    {
	"Legendary",
	"Rare",
	"Uncommon",
	"Common",
	"Vermin",
	0
    };

    choice = glbEngine->askQuestionListCanned("query", "mob_rarity", choices, 0, isclient);
    if (choice < 0)
	return;

    int		rarity = 100;

    switch (choice)
    {
	case 0:
	    rarity = 1;
	    break;
	case 1:
	    rarity = 10;
	    break;
	case 2:
	    rarity = 50;
	    break;
	case 3:
	    rarity = 100;
	    break;
	case 4:
	    rarity = 200;
	    break;
    }

    mob->edefn().set_rarity = true;
    mob->edefn().rarity = rarity;
}

static bool
requestMobDamageReduction(const char *name, int level, int &reduction, bool isclient)
{
    int		choice;
    const char *choices[] =
    {
	"Takes Extra Damage",
	"None",
	"Minor Reduction",
	"Major Reduction",
	"Massive Reduction",
	0
    };

    choice = glbEngine->askQuestionListCanned("query", "mob_damagereduction", choices, 1, isclient);
    if (choice < 0)
	return false;

    reduction = 0;
    switch (choice)
    {
	case 0:
	    reduction = -rand_range((level+1)/2, level);
	    break;
	case 1:
	    reduction = 0;
	    break;
	case 2:
	    reduction = rand_range((level+1)/2, level);
	    break;
	case 3:
	    reduction = rand_range(level, level+level/2);
	    break;
	case 4:
	    reduction = rand_range(level+level/2, 2*level);
	    break;
	case 5:
	    reduction = rand_range(2*level, 3*level);
	    break;
    }

    return true;
}

static bool
requestMobDodgeBonus(const char *name, int &bonus, bool isclient)
{
    int		choice;
    const char *choices[] =
    {
	"Steps into Attack",
	"Stumbles",
	"None",
	"Slippery",
	"Leaps Lightly Away",
	0
    };

    choice = glbEngine->askQuestionListCanned("query", "mob_dodgebonus", choices, 2, isclient);
    if (choice < 0)
	return false;

    bonus = 0;
    switch (choice)
    {
	case 0:
	    bonus = rand_range(-6, -3);
	    break;
	case 1:
	    bonus = rand_range(-2, -1);
	    break;
	case 2:
	    bonus = 0;
	    break;
	case 3:
	    bonus = rand_range(1, 2);
	    break;
	case 4:
	    bonus = rand_range(3, 6);
	    break;
    }

    return true;
}

void
defedit_editMobArmour(MOB *mob, bool isclient)
{
    int		reduction = 0;
    bool	setreduce = false;
    bool	setdodge = false;

    mob->setupTextVariables();

    if (requestMobDamageReduction(mob->getName(), mob->getLevel(), reduction, isclient))
    {
	setreduce = true;

	mob->edefn().damagereduction = reduction;
    }

    if (requestMobDodgeBonus(mob->getName(), reduction, isclient))
    {
	setdodge = true;

	mob->edefn().dodgebonus = reduction;
    }

    if (setreduce && setdodge)
    {
	mob->edefn().set_damagereduction = true;
    }
}

static bool
requestItemDamageReduction(const char *name, int level, int &reduction, bool isclient)
{
    int		choice;
    const char *choices[] =
    {
	"Less than Normal",
	"Normal",
	"Extra Absorbption",
	"Massive Absorbption",
	0
    };

    choice = glbEngine->askQuestionListCanned("query", "item_damagereduction", choices, 1, isclient);
    if (choice < 0)
	return false;

    reduction = 0;
    switch (choice)
    {
	case 0:
	    reduction = rand_range((level+1)/2, level);
	    break;
	case 1:
	    reduction = rand_range(level, level+level/2);
	    break;
	case 2:
	    reduction = rand_range(level+level/2, 2*level);
	    break;
	case 3:
	    reduction = rand_range(2*level, 3*level);
	    break;
    }

    return true;
}

static bool
requestItemDamageReduction(BUF name, int level, int &reduction, bool isclient)
{
    return requestItemDamageReduction(name.buffer(), level, reduction, isclient);
}
void
defedit_editItemArmour(ITEM *item, bool isclient)
{
    int		reduction = 0;

    item->setupTextVariables();

    if (requestItemDamageReduction(item->getName(), item->getDepth(), reduction, isclient))
    {
	item->edefn().set_damagereduction = true;
	item->edefn().damagereduction = reduction;
    }
}

void
defedit_editAttackDamage(MOB *mob, ITEM *item, ATTACK_NAMES attack, bool isclient)
{
    const char		*key = mob ? "mob_attackdamage" : "item_attackdamage";

    int		hitcount = 0;
    int		level = 1;

    if (mob) mob->setupTextVariables();
    if (item) item->setupTextVariables();

    if (mob) level = mob->getLevel();
    if (item) level = item->getDepth();

    const char		*choices[] =
    {
	"Pitiful",
	"Weak",
	"Normal",
	"Strong",
	"Insta-kill",
	0
    };

    int choice = glbEngine->askQuestionListCanned("query", key, choices, 2, isclient);

    // Abort...
    if (choice < 0)
	return;

    // Possibly make a new attack.
    if (attack == ATTACK_NONE)
    {
	attack = GAMEDEF::createNewAttackDef();
	if (mob)
	    mob->edefn().melee_attack = attack;
	if (item)
	    item->edefn().melee_attack = attack;
    }
    ATTACK_DEF	*def = GAMEDEF::attackdef(attack);

    float		hitstokill = 10;
    switch (choice)
    {
	case 0:
	    hitstokill = (float)rand_range(25, 30);
	    if (item)
		hitstokill /= 6;		// 4.16..5
	    break;
	case 1:
	    hitstokill = (float)rand_range(15, 20);
	    if (item)
		hitstokill /= 5;		// 3..4
	    break;
	case 2:
	    hitstokill = (float)rand_range(10, 15);
	    if (item)
		hitstokill /= 5;		// 2..3
	    break;
	case 3:
	    hitstokill = (float)rand_range(8, 12);
	    if (item)
		hitstokill /= 5;		// 1.6..2.4
	    break;
	case 4:
	    hitstokill = (float)rand_range(5, 8);
	    if (item)
		hitstokill /= 4;		// 1.25..2
	    break;
    }

    if (item && item->isRanged())
    {
	// Ranged weapons must be much lower.
	hitstokill *= 5;
    }

    float damage = MOB::expectedHPforLevel(level) / hitstokill;

    int	hitdamage = (int) (damage + 0.5);
    if (hitdamage < 1)
	hitdamage = 1;
    def->damage = hitdamage;
}

void
defedit_editAttack(MOB *mob, ITEM *item, ATTACK_NAMES attack, bool isclient)
{
    if (attack == ATTACK_NONE)
    {
	attack = GAMEDEF::createNewAttackDef();
	if (mob)
	    mob->edefn().melee_attack = attack;
	if (item)
	    item->edefn().melee_attack = attack;
    }

    // Now send in a generic edit...
    if (mob) mob->setupTextVariables();
    if (item) item->setupTextVariables();

    int			level = 1;
    if (mob) level = mob->getLevel();
    if (item) level = item->getDepth();

    ENGINE::Question		query;
    query.query = text_lookup("query", mob ? "mob_attackedit" : "item_attackedit");
    query.type = ENGINE::QUERY_ATTACK;
    query.payload = attack;
    query.mob = mob;
    query.item = item;

    glbEngine->askQuestion(query, isclient);
}

void
defedit_editAttackDistribution(MOB *mob, ITEM *item, ATTACK_NAMES attack, bool isclient)
{
    int		choice;

    text_registervariable("ATTACKNAME", GAMEDEF::attackdef(attack)->noun);

    PTRLIST<BUF>	choices;
    DISTRIBUTION_NAMES	dist;
    FOREACH_DISTRIBUTION(dist)
    {
	BUF	tmp;
	tmp.reference(glb_distributiondefs[dist].name);
	choices.append(tmp);
    }

    choice = glbEngine->askQuestionListCanned("query", "attack_distribution", choices, 2, isclient);
    if (choice < 0)
	return;

    GAMEDEF::attackdef(attack)->distribution = choice;
}

void
defedit_editAttackBonusToHit(MOB *mob, ITEM *item, ATTACK_NAMES attack, bool isclient)
{
    int		choice;

    text_registervariable("ATTACKNAME", GAMEDEF::attackdef(attack)->noun);

    const char *choices[] =
    {
	"Misses wildly",
	"Slighlty off balance",
	"Normal",
	"Well balanced",
	"Heat seeking",
	0
    };

    choice = glbEngine->askQuestionListCanned("query", "attack_bonustohit", choices, 2, isclient);
    if (choice < 0)
	return;

    int	results[] =
    {
	-3,
	-1,
	0,
	1,
	3
    };

    GAMEDEF::attackdef(attack)->chancebonus = results[choice];
}

ELEMENT_NAMES
defedit_chooseElement(ELEMENT_NAMES defelem, bool isclient)
{
    int		choice;
    int		defchoice = 0;

    PTRLIST<BUF>	choices;
    PTRLIST<ELEMENT_NAMES>	elems;
    BUF	tmp;

    while (1)
    {
	choices.clear();
	elems.clear();
	for (int e = 0; e < GAMEDEF::getNumElement(); e++)
	{
	    if (e == ELEMENT_NONE)
		continue;

	    if (e == defelem)
		defchoice = choices.entries();
	    tmp.reference(GAMEDEF::elementdef(e)->name);
	    choices.append(tmp);
	    elems.append((ELEMENT_NAMES) e);
	}

	tmp.reference("Edit Elements");
	choices.append(tmp);



	choice = glbEngine->askQuestionListCanned("query", "choose_element", choices, defchoice, isclient);
	if (choice < 0)
	    return ELEMENT_NONE;

	if (choice == choices.entries()-1)
	{
	    // Edit elements...
	    editElements(isclient);
	}
	else
	{
	    // Select!
	    return elems(choice);
	}
    }

    return ELEMENT_NONE;
}

void
defedit_editMobCorpseChance(MOB *mob, bool isclient)
{
    int		choice;

    mob->setupTextVariables();

    const char *choices[] =
    {
	"Never",
	"Sometimes",
	"50-50",
	"Usually",
	"Always",
	0
    };

    choice = glbEngine->askQuestionListCanned("query", "mob_corpsechance", choices, 2, isclient);
    if (choice < 0)
	return;

    int	results[] =
    {
	0,
	25,
	50,
	75,
	100
    };

    mob->edefn().set_corpsechance = true;
    mob->edefn().corpsechance = results[choice];
}

void
defedit_editMobLoot(MOB *mob, bool isclient)
{
    int		choice;

    mob->setupTextVariables();

    const char *choices[] =
    {
	"Nothing",
	"A paltry sum",
	"Normal",
	"Double",
	"Payday!",
	0
    };

    choice = glbEngine->askQuestionListCanned("query", "mob_loot", choices, 2, isclient);
    if (choice < 0)
	return;

    int	results[] =
    {
	0,
	1,
	2,
	4,
	10
    };

    mob->edefn().loot = mob->getLevel() * results[choice] / 2;
}

void
defedit_editMobSpecial(MOB *mob, bool isclient)
{
    mob->setupTextVariables();

    const char *choices[] =
    {
	"Change resistance",
	"Change vulnerability",
	"Toggle slow",
	"Toggle fast",
	"Toggle vampiric",
	"Toggle regenerate",
	"Toggle dig",
	"Toggle engulf",
	"Toggle breeder",
	"Toggle thief",
	"Toggle jumps",
	"Done",
	0
    };

    const int lastchoice = 11;
    int choice = lastchoice;

    BUF		prequel = text_lookup("query", "mob_special");

    while (1)
    {
	BUF	question = prequel;
	BUF	special = MOB::buildSpecialDescr(mob->getDefinition());

	if (special.isstring())
	{
	    question.strcat("Special: ");
	    question.strcat(special);
	}
	choice = glbEngine->askQuestionList(question, choices, choice, isclient);
	if (choice < 0)
	    return;

	mob->edefn().set_special = true;

	if (choice == lastchoice)
	    return;

	switch (choice)
	{
	    case 0:
	    {
		ELEMENT_NAMES	elem = defedit_chooseElement((ELEMENT_NAMES) mob->defn().resistance, isclient);
		mob->edefn().resistance = elem;
		break;
	    }
	    case 1:
	    {
		ELEMENT_NAMES	elem = defedit_chooseElement((ELEMENT_NAMES) mob->defn().vulnerability, isclient);
		mob->edefn().vulnerability = elem;
		break;
	    }
	    case 2:
		mob->edefn().isslow = !mob->edefn().isslow;
		break;
	    case 3:
		mob->edefn().isfast = !mob->edefn().isfast;
		break;
	    case 4:
		mob->edefn().isvampire = !mob->edefn().isvampire;
		break;
	    case 5:
		mob->edefn().isregenerate = !mob->edefn().isregenerate;
		break;
	    case 6:
		mob->edefn().candig = !mob->edefn().candig;
		break;
	    case 7:
		mob->edefn().swallows = !mob->edefn().swallows;
		break;
	    case 8:
		mob->edefn().breeder = !mob->edefn().breeder;
		break;
	    case 9:
		mob->edefn().isthief = !mob->edefn().isthief;
		break;
	    case 10:
		mob->edefn().canleap = !mob->edefn().canleap;
		break;
	}
    }
}

float
defedit_chooseAmount(const char *prequel, bool isclient)
{
    int		choice = 2;
    const char *choices[] =
    {
	"Negligible amount",
	"Below average",
	"Average",
	"Above average",
	"Huge amount",
	0
    };

    BUF	buf;
    buf.strcpy(prequel);
    choice = glbEngine->askQuestionList(buf, choices, choice, isclient);
    if (choice < 0)
	return 1.0;

    const float results[] =
    {
	0.1,
	0.5,
	1.0,
	1.5,
	2.0
    };

    return results[choice];
}

float
defedit_chooseDuration(const char *prequel, bool isclient)
{
    int		choice = 2;
    const char *choices[] =
    {
	"Just a second",
	"Just a minute",
	"Normal",
	"Long time",
	"Quite a long time",
	"Almost forever",
	0
    };

    BUF buf;
    buf.strcpy(prequel);
    choice = glbEngine->askQuestionList(buf, choices, choice, isclient);
    if (choice < 0)
	return 1.0;

    const float results[] =
    {
	0.1,
	0.5,
	1.0,
	2.0,
	4.0,
	30.0
    };

    return results[choice];
}

int
defedit_chooseAreaEffect(bool isclient)
{
    int		choice = 0;
    const char *choices[] =
    {
	"Single target",
	"Small blast",
	"Large blast",
	"Huge blast",
	0
    };

    BUF buf;
    buf.reference("How big of an area do you wish to affect?");
    choice = glbEngine->askQuestionList(buf, choices, choice, isclient);
    if (choice < 0)
	return 1;

    const int results[] =
    {
	1,
	2,
	3,
	4,
    };

    return results[choice];
}

int
defedit_chooseRange(bool isclient)
{
    int		choice = 2;
    const char *choices[] =
    {
	"None",
	"Short",
	"Normal",
	"Long",
	0
    };

    BUF buf;
    buf.reference("How long of a range?");
    choice = glbEngine->askQuestionList(buf, choices, choice, isclient);
    if (choice < 0)
	return 5;

    const int results[] =
    {
	0,
	3,
	6,
	10,
    };

    return results[choice];
}

int
defedit_chooseRecharge(bool isclient)
{
    int		choice = 1;
    const char *choices[] =
    {
	"Instant",
	"Brief",
	"Normal",
	"Long",
	"Almost forever",
	0
    };

    BUF buf;
    buf.reference("How long to recharge?");
    choice = glbEngine->askQuestionList(buf, choices, choice, isclient);
    if (choice < 0)
	return 5;

    const int results[] =
    {
	0,
	2,
	5,
	10,
	100,
    };

    return results[choice];
}

EFFECT_NAMES
defedit_editEffect(EFFECT_NAMES origeffect, int level, bool isclient)
{
    BUF		blank;

    return defedit_editEffect(blank, origeffect, level, isclient);
}

EFFECT_NAMES
defedit_editEffect(BUF prequel, EFFECT_NAMES origeffect, int level, bool isclient)
{
    const int		lastchoice = 11;
    int		choice = lastchoice;

    const char *choices[] =
    {
	"Nothing",
	"Heal",
	"Damage",
	"Gain mana",
	"Lose mana",
	"Gain food",
	"Lose food",
	"Poison",
	"Cure",
	"Resist element",
	"Vulnerable to element",
	"Done",
	0
    };

    while (1)
    {
	BUF		query;
	query = prequel;
	if (query.isstring())
	    query.strcat("\n\n");
	query.strcat(text_lookup("query", "edit_effect"));
	BUF		stat = MOB::buildEffectDescr(origeffect);
	query.strcat("Current Effect: ");
	if (stat.isstring())
	    query.strcat(stat);
	else
	    query.strcat("None");

	choice = glbEngine->askQuestionList(query, choices, choice, isclient);
	if (choice < 0)
	    break;
	if (choice == lastchoice)
	    break;

	// A real change, possible make a new effect.
	if (origeffect == EFFECT_NONE)
	    origeffect = GAMEDEF::createNewEffectDef();

	EFFECT_DEF	*def = GAMEDEF::effectdef(origeffect);
	float		 scale = 1.0;

	switch (choice)
	{
	    case 0:
		def->type = EFFECTCLASS_NONE;
		break;
	    case 1:
		// Heal
		def->type = EFFECTCLASS_HEAL;
		scale = defedit_chooseAmount("How much will this heal?", isclient);

		def->amount = MOB::expectedHPforLevel(level) * (scale * 0.5);
		def->distribution = defedit_chooseDistribution("How consistent are these heals?", (DISTRIBUTION_NAMES) def->distribution, isclient);
		break;
	    case 2:
		// Damage...
		def->type = EFFECTCLASS_DAMAGE;
		scale = defedit_chooseAmount("How much will this hurt?", isclient);

		def->amount = MOB::expectedHPforLevel(level) * (scale * 0.2);
		def->distribution = defedit_chooseDistribution("How consistent is the damage?", (DISTRIBUTION_NAMES) def->distribution, isclient);
		if (def->element == ELEMENT_NONE)
		    def->element = ELEMENT_PHYSICAL;
		def->element = defedit_chooseElement((ELEMENT_NAMES) def->element, isclient);
		break;
	    case 3:
		def->type = EFFECTCLASS_GAINMANA;
		scale = defedit_chooseAmount("How much mana will this restore?", isclient);

		def->amount = MOB::expectedMPforLevel(level) * (scale * 0.5);
		def->distribution = defedit_chooseDistribution("How consistent is the restoration?", (DISTRIBUTION_NAMES) def->distribution, isclient);
		break;
	    case 4:
		def->type = EFFECTCLASS_LOSEMANA;
		scale = defedit_chooseAmount("How much mana will this drain?", isclient);

		def->amount = MOB::expectedMPforLevel(level) * (scale * 0.2);
		def->distribution = defedit_chooseDistribution("How consistent is the draining?", (DISTRIBUTION_NAMES) def->distribution, isclient);
		break;
	    case 5:
		def->type = EFFECTCLASS_GAINFOOD;
		scale = defedit_chooseAmount("How much food will this create?", isclient);
		def->amount = 100 * (scale);
		def->distribution = defedit_chooseDistribution("How consistent is the creation?", (DISTRIBUTION_NAMES) def->distribution, isclient);
		break;
	    case 6:
		def->type = EFFECTCLASS_LOSEFOOD;
		scale = defedit_chooseAmount("How much food will this consume?", isclient);
		def->amount = 100 * (scale);
		def->distribution = defedit_chooseDistribution("How consistent is the consumption?", (DISTRIBUTION_NAMES) def->distribution, isclient);
		break;
	    case 7:
		/// Poison
		def->type = EFFECTCLASS_POISON;
		scale = defedit_chooseDuration("How long will the poison last?", isclient);
		def->duration = scale * 10;
		break;
	    case 8:
		// Cure
		def->type = EFFECTCLASS_CURE;
		break;
	    case 9:
		def->type = EFFECTCLASS_RESIST;
		scale = defedit_chooseDuration("How long will the resistance last?", isclient);
		def->duration = scale * 50;
		def->element = defedit_chooseElement((ELEMENT_NAMES) def->element, isclient);
		break;
	    case 10:
		def->type = EFFECTCLASS_VULNERABLE;
		scale = defedit_chooseDuration("How long will the vulnerability last?", isclient);
		def->duration = scale * 50;
		def->element = defedit_chooseElement((ELEMENT_NAMES) def->element, isclient);
		break;
	}
    }

    return origeffect;
}

DISTRIBUTION_NAMES
defedit_chooseDistribution(const char *prequel, DISTRIBUTION_NAMES origdist, bool isclient)
{
    int		choice;

    PTRLIST<BUF>	choices;
    DISTRIBUTION_NAMES	dist;
    FOREACH_DISTRIBUTION(dist)
    {
	BUF	tmp;
	tmp.reference(glb_distributiondefs[dist].name);
	choices.append(tmp);
    }

    BUF buf;
    buf.strcpy(prequel);
    choice = glbEngine->askQuestionList(buf, choices, origdist, isclient);
    if (choice < 0)
	return origdist;

    return (DISTRIBUTION_NAMES) choice;
}

void
defedit_editRangeAttack(MOB *mob, bool isclient)
{
    ENGINE::Question	query;
    query.type = ENGINE::QUERY_RANGEATTACK;
    query.mob = mob;

    glbEngine->askQuestion(query, isclient);
}
