/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	Jacob's Matrix Development
 *
 * NAME:        main.cpp ( Jacob's Matrix, C++ )
 *
 * COMMENTS:
 */

#include <libtcod.hpp>
#include <stdio.h>
#include <time.h>

//#define USE_AUDIO
//#define JOY_SUPPORT

#define POPUP_DELAY 1000

#ifdef WIN32
#include <windows.h>
#include <process.h>
#define usleep(x) Sleep((x)*1000)
// #pragma comment(lib, "SDL.lib")
// #pragma comment(lib, "SDLmain.lib")

// I really hate windows.
#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif

#else
#include <unistd.h>
#endif

#include <fstream>
using namespace std;

int		 glbLevelStartMS = -1;

#include "rand.h"
#include "gfxengine.h"
#include "config.h"
#include "builder.h"
#include "defedit.h"
#include "dircontrol.h"

// #include <SDL.h>
#ifdef USE_AUDIO
#include <SDL_mixer.h>

Mix_Music		*glbTunes = 0;
#endif
bool			 glbMusicActive = false;

// Cannot be bool!
// libtcod does not accept bools properly.

volatile int		glbItemCache = -1;

SPELL_NAMES		glbLastSpell = SPELL_NONE;
int			glbLastItemIdx = 0;
int			glbLastItemAction = 0;
const char		*glbWorldName = "Default";
bool			glbReloadedLevel = true;

void endOfMusic()
{
    // Keep playing.
    // glbMusicActive = false;
}

void
setMusicVolume(int volume)
{
#ifdef USE_AUDIO
    volume = BOUND(volume, 0, 10);

    glbConfig->myMusicVolume = volume;

    Mix_VolumeMusic((MIX_MAX_VOLUME * volume) / 10);
#endif
}

void
stopMusic(bool atoncedamnit = false)
{
#ifdef USE_AUDIO
    if (glbMusicActive)
    {
	if (atoncedamnit)
	    Mix_HaltMusic();
	else
	    Mix_FadeOutMusic(2000);
	glbMusicActive = false;
    }
#endif
}

void
startMusic()
{
#ifdef USE_AUDIO
    if (glbMusicActive)
	stopMusic();

    if (!glbTunes)
	glbTunes = Mix_LoadMUS(glbConfig->musicFile());
    if (!glbTunes)
    {
	printf("Failed to load %s, error %s\n", 
		glbConfig->musicFile(), 
		Mix_GetError());
    }
    else
    {
	glbMusicActive = true;
	glbLevelStartMS = TCOD_sys_elapsed_milli();

	if (glbConfig->musicEnable())
	{
	    Mix_PlayMusic(glbTunes, -1);		// Infinite
	    Mix_HookMusicFinished(endOfMusic);

	    setMusicVolume(glbConfig->myMusicVolume);
	}
    }
#endif
}


#ifdef main
#undef main
#endif

#include "mob.h"
#include "map.h"
#include "text.h"
#include "display.h"
#include "engine.h"
#include "panel.h"
#include "msg.h"
#include "firefly.h"
#include "item.h"
#include "chooser.h"

PANEL		*glbMessagePanel = 0;
DISPLAY		*glbDisplay = 0;
MAP		*glbMap = 0;
ENGINE		*glbEngine = 0;
PANEL		*glbPopUp = 0;
PANEL		*glbJacob = 0;
PANEL		*glbWeaponInfo = 0;
PANEL		*glbVictoryInfo = 0;
bool		 glbPopUpActive = false;
bool		 glbRoleActive = false;
FIREFLY		*glbHealthFire = 0;
FIREFLY		*glbManaFire = 0;
FIREFLY		*glbFoodFire = 0;
FIREFLY		*glbCoinStack = 0;
int		 glbLevel = 1;
int		 glbMaxLevel = 0;
PTRLIST<int>	 glbLevelTimes;
bool		 glbVeryFirstRun = true;

CHOOSER		*glbLevelBuildStatus = 0;

CHOOSER		*glbChooser = 0;
bool		 glbChooserActive = false;

int		 glbInvertX = -1, glbInvertY = -1;
int		 glbMeditateX = -1, glbMeditateY = -1;

#define DEATHTIME 15000


BUF
mstotime(int ms)
{
    int		sec, min;
    BUF		buf;

    sec = ms / 1000;
    ms -= sec * 1000;
    min = sec / 60;
    sec -= min * 60;

    if (min)
	buf.sprintf("%dm%d.%03ds", min, sec, ms);
    else if (sec)
	buf.sprintf("%d.%03ds", sec, ms);
    else
	buf.sprintf("0.%03ds", ms);

    return buf;
}

class MOBPTR_SORT
{
public:
    bool operator()(MOB *a, MOB *b)
    {
	return a->getUID() < b->getUID();
    }
};

class ITEMPTR_SORT
{
public:
    bool operator()(ITEM *a, ITEM *b)
    {
	return a->getUID() < b->getUID();
    }
};

class ITEMPTR_QUALITYSORT
{
public:
    bool operator()(ITEM *a, ITEM *b)
    {
	if (a->isBroken() && !b->isBroken())
	    return true;
	if (!a->isBroken() && b->isBroken())
	    return false;
	if (a->isMelee())
	{
	    if (b->isArmour())
		return true;
	    if (b->isMelee())
		return (a->attackdefn().damage < b->attackdefn().damage);
	    return true;
	}
	if (a->isArmour())
	{
	    if (b->isMelee())
		return false;
	    if (b->isArmour())
		return (a->getDamageReduction() < b->getDamageReduction());
	    return true;
	}
	return a->getUID() < b->getUID();
    }
};

// Should call this in all of our loops.
void
redrawWorldNoFlush()
{
    static MAP	*centermap = 0;
    static POS	 mapcenter;
    bool	 isblind = false;
    bool	 isvictory = false;

    gfx_updatepulsetime();

    // Wipe the world
    for (int y = 0; y < SCR_HEIGHT; y++)
	for (int x = 0; x < SCR_WIDTH; x++)
	    gfx_printchar(x, y, ' ');

    if (glbMap)
	glbMap->decRef();
    glbMap = glbEngine ? glbEngine->copyMap() : 0;

    glbWeaponInfo->clear();

    glbMeditateX = -1;
    glbMeditateY = -1;
    int			avataralive = 0;
    if (glbMap && glbMap->avatar())
    {
	MOB		*avatar = glbMap->avatar();
	int		 timems;

	avataralive = avatar->alive();

	if (avatar->alive() && avatar->isMeditating())
	{
	    glbMeditateX = glbDisplay->x() + glbDisplay->width()/2;
	    glbMeditateY = glbDisplay->y() + glbDisplay->height()/2;
	}

	timems = TCOD_sys_elapsed_milli();

	if (centermap)
	    centermap->decRef();
	centermap = glbMap;
	centermap->incRef();
	mapcenter = avatar->meditatePos();

	isblind = avatar->hasItem(ITEM_BLIND);

	if (avatar->hasItem(ITEM_INVULNERABLE))
	    glbHealthFire->setFireType(FIRE_MONO);
	else if (avatar->hasItem(ITEM_POISON))
	    glbHealthFire->setFireType(FIRE_POISON);
	else
	    glbHealthFire->setFireType(FIRE_HEALTH);

	glbHealthFire->setRatioLiving(avatar->getHP() ? (avatar->getHP() / (float) avatar->getMaxHP()) : 0);
	glbManaFire->setRatioLiving(avatar->getMP() ? (avatar->getMP() / (float) avatar->getMaxMP()) : 0);
	glbFoodFire->setRatioLiving(avatar->getFood() ? avatar->getFood() / (float) avatar->getMaxFood() : 0);
	{
	    int		ngold = 0;
	    ITEM	*gold = avatar->lookupItem(ITEM_GOLD);
	    if (gold)
		ngold = gold->getStackCount();

	    glbCoinStack->setParticleCount(BOUND(ngold, 0, 22 * SCR_HEIGHT));
	    glbCoinStack->setBloodCoins(0);
	    glbCoinStack->setRatioLiving((float) (BOUND(ngold, 0, 20*(SCR_HEIGHT-2)) / float(20*(SCR_HEIGHT-2))));
	}

	BUF		buf;

	{
	    MOBLIST	atlist;

	    glbWeaponInfo->setTextAttr(ATTR_WHITE);
	    glbWeaponInfo->clear();
	}
    }
    
    glbDisplay->display(mapcenter, isblind);
    glbMessagePanel->redraw();

    {
	glbJacob->clear();

	if (glbMap)
	{
	    glbJacob->setTextAttr(ATTR_NORMAL);
	    glbJacob->appendText("\n");
	}
	

	if (glbMap && glbMap->avatar())
	{
	    MOB	*avatar = glbMap->avatar();
	    BUF		buf;

	    glbJacob->setTextAttr(ATTR_WHITE);
	    buf.sprintf("Depth: %d\n", glbMap->getDepth());
	    glbJacob->appendText(buf);

	    if (glbMap->avatar()->isSwallowed())
	    {
		glbJacob->setTextAttr(ATTR_YELLOW);
		glbJacob->appendText("Swallowed!\n");
	    }
	    if (!glbMap->avatar()->getFood())
	    {
		glbJacob->setTextAttr(ATTR_YELLOW);
		glbJacob->appendText("Starving!\n");
	    }
	    if (glbMap->avatar()->hasItem(ITEM_POISON))
	    {
		glbJacob->setTextAttr(ATTR_GREEN);
		glbJacob->appendText("Poisoned!\n");
	    }

	    glbJacob->setTextAttr(ATTR_WHITE);
	    glbJacob->appendText("\nInventory:\n");

	    glbJacob->setTextAttr(ATTR_NORMAL);
	    if (!avatar->inventory().entries())
	    {
		glbJacob->appendText("  Nothing\n");
	    }
	    ITEMLIST	items = avatar->inventory();

	    ITEMPTR_SORT	op;

	    items.stableSort(op);

	    for (int i = 0; i < items.entries(); i++)
	    {
		BUF	itemname = items(i)->getName();
		itemname = gram_capitalize(itemname);
		if (items(i)->isEquipped() && items(i)->isMelee())
		    buf.sprintf("  %s (held)\n", itemname.buffer());
		else if (items(i)->isEquipped() && items(i)->isArmour())
		    buf.sprintf("  %s (worn)\n", itemname.buffer());
		else
		    buf.sprintf("  %s\n", itemname.buffer());
		glbJacob->appendText(buf);
	    }
	}
    }

    glbJacob->redraw();

    glbWeaponInfo->redraw();

    if (isvictory)
	glbVictoryInfo->redraw();

    {
	// we have to clear the RHS.
	static FIRETEX		 *cleartex = 0;
	if (!cleartex)
	{
	    cleartex = new FIRETEX(20, (SCR_HEIGHT-2));
	    cleartex->buildFromConstant(false, 0, FIRE_HEALTH);
	}
	cleartex->redraw(60, 2);
    }
    if (glbCoinStack)
    {
	FIRETEX		*tex = glbCoinStack->getTex();
	tex->redraw(0, 0);
	tex->decRef();
    }
    if (glbHealthFire)
    {
	FIRETEX		*tex = glbHealthFire->getTex();
	tex->redraw(SCR_WIDTH-3, 0);
	tex->decRef();
    }
    if (glbManaFire)
    {
	FIRETEX		*tex = glbManaFire->getTex();
	tex->redraw(SCR_WIDTH-2, 0);
	tex->decRef();
    }
    if (glbFoodFire)
    {
	FIRETEX		*tex = glbFoodFire->getTex();
	tex->redraw(SCR_WIDTH-1, 0);
	tex->decRef();
    }

    if (glbMeditateX >= 0 && glbMeditateY >= 0)
    {
	gfx_printattrback(glbMeditateX, glbMeditateY, ATTR_AVATARMEDITATE);
    }

    if (glbInvertX >= 0 && glbInvertY >= 0)
    {
	gfx_printattr(glbInvertX, glbInvertY, ATTR_HILITE);
    }


    if (glbItemCache >= 0)
    {
	glbLevelBuildStatus->clear();

	glbLevelBuildStatus->appendChoice("   Building Distance Caches   ", glbItemCache);
	glbLevelBuildStatus->redraw();
    }

    if (glbChooserActive)
    {
	glbChooser->redraw();
    }

    if (glbPopUpActive)
	glbPopUp->redraw();
}

void
redrawWorld()
{
    redrawWorldNoFlush();
    TCODConsole::flush();
}

// Pops up the given text onto the screen.
// The key used to dismiss the text is the return code.
int
popupText(const char *buf, int delayms = 0)
{
    int		key = 0;
    int		startms = TCOD_sys_elapsed_milli();

    glbPopUp->clear();
    glbPopUp->resize(50, SCR_HEIGHT-2);
    glbPopUp->move(10, 1);

    // A header matches the footer.
    glbPopUp->appendText("\n");

    // Must make active first because append text may trigger a 
    // redraw if it is long!
    glbPopUpActive = true;
    glbPopUp->appendText(buf);

    // In case we drew due to too long text
    glbPopUp->erase();

    // Determine the size of the popup and resize.
    glbPopUp->shrinkToFit();

    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();
	key = gfx_getKey(false);
	if (delayms)
	{
	    int		currentms = (int)TCOD_sys_elapsed_milli() - startms;
	    float	ratio = currentms / (float)delayms;
	    ratio = BOUND(ratio, 0, 1);

	    glbPopUp->fillRect(0, glbPopUp->height()-1, (int)(glbPopUp->width()*ratio+0.5), 1, ATTR_WAITBAR);
	}
	if (key)
	{
	    // Don't let them abort too soon.
	    if (((int)TCOD_sys_elapsed_milli() - startms) >= delayms)
		break;
	}
    }

    glbPopUpActive = false;
    glbPopUp->erase();

    glbPopUp->resize(50, 15);
    glbPopUp->move(15, 5);

    redrawWorld();

    return key;
}

int
popupText(BUF buf, int delayms = 0)
{
    return popupText(buf.buffer(), delayms);
}

// Pops up the given text onto the screen.
// Allows the user to enter a response.
// Returns what they typed.
BUF
popupQuestion(const char *buf)
{
#define RESPONSE_LEN 50
    char	response[RESPONSE_LEN];

    glbPopUp->clear();
    glbPopUp->resize(50, SCR_HEIGHT-2);
    glbPopUp->move(15, 1);

    // A header matches the footer.
    glbPopUp->appendText("\n");

    // Must make active first because append text may trigger a 
    // redraw if it is long!
    glbPopUpActive = true;
    glbPopUp->appendText(buf);
    glbPopUp->appendText("\n");

    // In case we drew due to too long text
    glbPopUp->erase();

    // Determine the size of the popup and resize.
    glbPopUp->shrinkToFit();
    glbPopUp->resize( MAX(RESPONSE_LEN + 2, glbPopUp->getW()),
		      glbPopUp->getH() + 3 );

    redrawWorld();
    glbPopUp->getString("\n> ", response, RESPONSE_LEN);

    BUF		result;
    result.strcpy(response);

    glbPopUpActive = false;
    glbPopUp->erase();

    glbPopUp->resize(50, 30);
    glbPopUp->move(15, 10);

    redrawWorld();

    return result;
}

BUF
popupQuestion(BUF buf)
{
    return popupQuestion(buf.buffer());
}

template <typename T>
T
runMenuSelection(const PTRLIST<T> &list, T fallback, 
		 const PTRLIST<int> &hotkeys)
{
    // Run the chooser...
    if (!list.entries())
	return fallback;

    int		key;
    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();
	key = gfx_getKey(false);

	glbChooser->processKey(key);

	if (key)
	{
	    int		itemno = glbChooser->getChoice();
	    // User selects this?
	    if (key == '\n' || key == ' ')
	    {
		fallback = list(itemno);
		break;
	    }
	    else if (key == '\x1b')
	    {
		// Esc on options is to go back to play.
		// Play...
		break;
	    }
	    else
	    {
		// Ignore other options.
		for (int i = 0; i < hotkeys.entries(); i++)
		{
		    if (key == hotkeys(i))
		    {
			fallback = list(i);
			glbChooserActive = false;
			return fallback;
		    }
		}
	    }
	}
    }
    glbChooserActive = false;
    return fallback;
}

//
// Prints the welcome text, with proper varialbes
//
void
welcomeText(const char *key)
{
    text_registervariable("ITEMNAME", GAMEDEF::itemdef(ITEM_MACGUFFIN)->name);
    text_registervariable("MOBNAME", GAMEDEF::mobdef(MOB_BOSS)->name);
    BUF		depth;
    depth.reference("unknown");
    if (GAMEDEF::rules().set_bosslevel)
    {
	depth.sprintf("%d", GAMEDEF::rules().bosslevel);
    }
    text_registervariable("DEPTH", depth);

    popupText(text_lookup("welcome", key));
}

//
// Freezes the game until a key is pressed.
// If key is not direction, false returned and it is eaten.
// else true returned with dx/dy set.  Can be 0,0.
bool
awaitDirection(int &dx, int &dy)
{
    int			key;

    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();

	key = gfx_getKey(false);

	if (gfx_cookDir(key, dx, dy))
	{
	    return true;
	}
	if (key)
	    return false;
    }

    return false;
}

void
editMob(MOB *mob)
{
    MOB_DEF	*def = GAMEDEF::mobdef(mob->getDefinition());

    const int		lastchoice = 10;
    int			choice = lastchoice;
    ENGINE::Question		query;

    const char *choices[] =
    {
	"Behaviour",
	"Abundance",
	"Health",
	"Natural Armour",
	"Melee Attack",
	"Range Attack",
	"Corpse Chance",
	"Corpse Effect",
	"Loot",
	"Special",
	"Done",
	0
    };

    query.setChoices(choices);
    query.query.sprintf("Adjust which properties of %s?",
		mob->getName());
    query.type = ENGINE::QUERY_CHOICE;

    while (!TCODConsole::isWindowClosed())
    {
	query.defaultchoice = choice;
	ENGINE::Response	response = defedit_askQuestion(query);

	if (response.type != ENGINE::QUERY_CHOICE)
	{
	    break;
	}

	choice = response.choice;

	if (choice == lastchoice)
	    break;
	switch (choice)
	{
	    case 0:
		defedit_editMobAI(mob, true);
		break;
	    case 1:
		defedit_editMobRarity(mob, true);
		break;
	    case 2:
		defedit_editMobHP(mob, true);
		break;
	    case 3:
		defedit_editMobArmour(mob, true);
		break;
	    case 4:
		defedit_editAttack(mob, 0, (ATTACK_NAMES)mob->defn().melee_attack, true);
		break;
	    case 5:
		defedit_editRangeAttack(mob, true);
		break;
	    case 6:
		defedit_editMobCorpseChance(mob, true);
		break;
	    case 7:
	    {
		BUF		prequel;
		BUF		name;
		name.strcat(gram_getarticle(mob->getName()));
		name.strcat(mob->getName());
		prequel.sprintf("What extra thing happens when you eat %s?", name.buffer());

		mob->edefn().corpse_effect =
		    defedit_editEffect(prequel, 
				(EFFECT_NAMES) mob->defn().corpse_effect,
				mob->getLevel(),
				true);
		break;
	    }
	    case 8:
		defedit_editMobLoot(mob, true);
		break;
	    case 9:
		defedit_editMobSpecial(mob, true);
		break;
	}
    }
}

void
editLook(const char *thing, BUF &name, u8 &symbol, ATTR_NAMES attr_name)
{
    ATTR_DEF	*attr = GAMEDEF::attrdef(attr_name);
    bool		done = false;

    int			choice = 9;

    while (!TCODConsole::isWindowClosed() && !done)
    {
	glbChooserActive = false;
	glbChooser->clear();

	glbChooser->setAttr(ATTR_NORMAL, ATTR_PERCENT_SELECT, ATTR_PERCENT_HILIGHT, ATTR_PERCENT_SELECTHILIGHT);
	glbChooser->setPrequelAttr(attr_name);
	BUF		buf;
	buf.sprintf("The %s %s is represented by %c.",
		    thing,
		    name.buffer(),
		    symbol);
	glbChooser->setPrequel(buf);

	glbChooser->appendChoice("Change Name");
	glbChooser->appendChoice("Change Symbol");
	glbChooser->appendChoice("Foreground Red", ((int)attr->fg_r) * 100 / 255);
	glbChooser->appendChoice("Foreground Green", ((int)attr->fg_g) * 100 / 255);
	glbChooser->appendChoice("Foreground Blue", ((int)attr->fg_b) * 100 / 255);
	glbChooser->appendChoice("Background Red", ((int)attr->bg_r) * 100 / 255);
	glbChooser->appendChoice("Background Green", ((int)attr->bg_g) * 100 / 255);
	glbChooser->appendChoice("Background Blue", ((int)attr->bg_b) * 100 / 255);
	if (!attr->pulse)
	    glbChooser->appendChoice("Turn On Pulse");
	else
	    glbChooser->appendChoice("Turn Off Pulse");
	glbChooser->appendChoice("Done");
    
	glbChooser->setChoice(choice);

	glbChooserActive = true;

	// Run the chooser...
	while (!TCODConsole::isWindowClosed())
	{
	    redrawWorld();
	    int key = gfx_getKey(false);

	    glbChooser->processKey(key);

	    int		dx, dy;

	    choice = glbChooser->getChoice();

	    if (gfx_cookDir(key, dx, dy, GFX_COOKDIR_X))
	    {
		u8		empty = 0;
		u8		*change = &empty;
		switch (choice)
		{
		    case 2:
			change = &attr->fg_r;
			break;
		    case 3:
			change = &attr->fg_g;
			break;
		    case 4:
			change = &attr->fg_b;
			break;
		    case 5:
			change = &attr->bg_r;
			break;
		    case 6:
			change = &attr->bg_g;
			break;
		    case 7:
			change = &attr->bg_b;
			break;
		}

		int		nval = *change;
		nval += 32 * dx;
		if (nval < 0) nval = 0;
		if (nval > 255) nval = 255;
		*change = nval;

		// Rebuild new percentages.
		break;
	    }

	    if (key)
	    {
		if (key == '\n' || key == ' ')
		{
		    if (choice == 9)
		    {
			// done!
			done = true;
			break;
		    }
		    if (choice == 0)
		    {
			BUF		rename, newname;
			rename.sprintf("What is a good new name for %s?",
					name.buffer());
			glbChooserActive = false;
			newname = popupQuestion(rename);

			if (newname.isstring())
			    name = newname;
			break;
		    }
		    if (choice == 1)
		    {
			// New symbol.
			BUF		rename, newname;
			rename.sprintf("What is a good new symbol for %s?",
					name.buffer());
			glbChooserActive = false;
			newname = popupQuestion(rename);

			if (newname.isstring())
			    symbol = newname.buffer()[0];
			break;
		    }
		    if (choice == 8)
		    {
			attr->pulse = !attr->pulse;
			break;
		    }
		}
		else if (key == '\x1b')
		{
		    // Esc on options is to go back to play.
		    // Play...
		    done = true;
		    break;
		}
	    }
	}

    }

    // REstore chooser
    glbChooser->setAttr(ATTR_NORMAL, ATTR_HILITE, ATTR_LIGHTBLACK, ATTR_DKHILITE);

    glbChooserActive = false;
}

void
displayMob(MOB *mob)
{
    bool		done = false;

    while (!TCODConsole::isWindowClosed() && !done)
    {
	// Display description with option to edit...
	glbChooser->clear();
	glbChooser->setTextAttr(ATTR_NORMAL);

	glbChooser->setPrequelAttr(ATTR_WHITE);

	BUF		longtext = mob->getLongDescription();
	longtext = longtext.wordwrap(SCR_WIDTH-15);
	longtext.stripWS();
	longtext.append('\n');
	longtext.append('\n');
	glbChooser->setPrequel(longtext);

	glbChooser->appendChoice("Edit Look");
	glbChooser->appendChoice("Edit Statistics");
	glbChooser->appendChoice("Done");

	glbChooser->setChoice(2);
	glbChooserActive = true;

	// Run the chooser...
	while (!TCODConsole::isWindowClosed())
	{
	    redrawWorld();
	    int key = gfx_getKey(false);

	    glbChooser->processKey(key);

	    if (key)
	    {
		if (key == '\n' || key == ' ')
		{
		    if (glbChooser->getChoice() == 2)
		    {
			// done!
			done = true;
			break;
		    }
		    else if (glbChooser->getChoice() == 1)
		    {
			glbChooserActive = false;

			editMob(mob);
			break;
		    }
		    else
		    {
			// Request to edit Look
			glbChooserActive = false;
			ATTR_NAMES	look = (ATTR_NAMES) mob->defn().attr;

			MOB_DEF		*mobdef = GAMEDEF::mobdef(mob->getDefinition());
			// Make a new look if needed.
			if (look < NUM_ATTRS)
			{
			    ATTR_NAMES	nlook = GAMEDEF::createNewAttrDef();
			    ATTR_DEF	*nattr = GAMEDEF::attrdef(nlook);
			    *nattr = *GAMEDEF::attrdef(look);
			    mobdef->attr = nlook;
			    look = nlook;
			}

			u8		symbol = mob->defn().symbol;
			BUF		name;
			name.reference(mob->defn().name);

			editLook("creature", name, symbol, look);

			if (name.strcmp(mobdef->name))
			    mobdef->name = glb_harden(name);
			mobdef->symbol = symbol;

			break;
		    }
		}
		else if (key == '\x1b')
		{
		    // Esc on options is to go back to play.
		    // Play...
		    done = true;
		    break;
		}
	    }
	}
    }
    glbChooserActive = false;
}

void
displayItem(ITEM *item)
{
    bool		done = false;

    while (!TCODConsole::isWindowClosed() && !done)
    {
	// Display description with option to edit...
	glbChooser->clear();
	glbChooser->setTextAttr(ATTR_NORMAL);

	glbChooser->setPrequelAttr(ATTR_WHITE);

	BUF		longtext = item->getLongDescription();
	longtext = longtext.wordwrap(SCR_WIDTH-15);
	longtext.stripWS();
	longtext.append('\n');
	longtext.append('\n');
	glbChooser->setPrequel(longtext);

	int		lastchoice = 1;

	glbChooser->appendChoice("Edit Look");
	if (item->isMelee() || item->isArmour() || item->isSpellbook() || item->isPotion())
	{
	    glbChooser->appendChoice("Edit Statistics");
	    lastchoice++;
	}
	glbChooser->appendChoice("Done");



	glbChooser->setChoice(lastchoice);
	glbChooserActive = true;

	// Run the chooser...
	while (!TCODConsole::isWindowClosed())
	{
	    redrawWorld();
	    int key = gfx_getKey(false);

	    glbChooser->processKey(key);

	    if (key)
	    {
		if (key == '\n' || key == ' ')
		{
		    if (glbChooser->getChoice() == lastchoice)
		    {
			// done!
			done = true;
			break;
		    }
		    else if (glbChooser->getChoice() == 1)
		    {
			glbChooserActive = false;
			if (item->isMelee())
			{
			    defedit_editAttack(0, item, (ATTACK_NAMES)item->defn().melee_attack, true);
			}
			if (item->isArmour())
			{
			    defedit_editItemArmour(item, true);
			}
			if (item->isPotion())
			{
			    item->edefn().potion_effect = defedit_editEffect((EFFECT_NAMES) item->defn().potion_effect, item->getDepth(), true);
			}
			if (item->isSpellbook())
			{
			    ENGINE::Question		question;
			    question.payload = item->defn().spell;
			    question.type = ENGINE::QUERY_SPELL;
			    glbEngine->askQuestion(question, true);
			}
			break;
		    }
		    else
		    {
			// Request to edit look...
			glbChooserActive = false;
			ATTR_NAMES	look = (ATTR_NAMES) item->defn().attr;

			ITEM_DEF		*itemdef = GAMEDEF::itemdef(item->getDefinition());
			// Make a new look if needed.
			if (look < NUM_ATTRS)
			{
			    ATTR_NAMES	nlook = GAMEDEF::createNewAttrDef();
			    ATTR_DEF	*nattr = GAMEDEF::attrdef(nlook);
			    *nattr = *GAMEDEF::attrdef(look);
			    itemdef->attr = nlook;
			    look = nlook;
			}

			u8		symbol = item->defn().symbol;
			BUF		name;
			name.reference(item->defn().name);

			editLook("item", name, symbol, look);

			if (name.strcmp(itemdef->name))
			    itemdef->name = glb_harden(name);
			itemdef->symbol = symbol;

			if (item->getDefinition() == ITEM_GOLD)
			{
			    ATTR_DEF	*def = GAMEDEF::attrdef(item->defn().attr);
			    glbCoinStack->setColor( def->fg_r, def->fg_g, def->fg_b );
			}

			break;
		    }
		}
		else if (key == '\x1b')
		{
		    // Esc on options is to go back to play.
		    // Play...
		    done = true;
		    break;
		}
	    }
	}
    }
    glbChooserActive = false;
}

void
doExamine()
{
    int		x = glbDisplay->x() + glbDisplay->width()/2;
    int		y = glbDisplay->y() + glbDisplay->height()/2;
    int		key;
    int		dx, dy;
    POS		p;
    BUF		buf;
    bool	blind;
    bool	lookclosely = false;

    glbInvertX = x;
    glbInvertY = y;
    blind = false;
    if (glbMap && glbMap->avatar())
	blind = glbMap->avatar()->hasItem(ITEM_BLIND);
    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();

	key = gfx_getKey(false);

	// We eat space below, so check here!
	if (key == ' ')
	{
	    lookclosely = true;
	    break;
	}

	if (gfx_cookDir(key, dx, dy))
	{
	    // Space we want to select.
	    if (!dx && !dy)
	    {
		break;
	    }
	    x += dx;
	    y += dy;
	    x = BOUND(x, glbDisplay->x(), glbDisplay->x()+glbDisplay->width()-1);
	    y = BOUND(y, glbDisplay->y(), glbDisplay->y()+glbDisplay->height()-1);

	    glbInvertX = x;
	    glbInvertY = y;
	    p = glbDisplay->lookup(x, y);

	    if (p.tile() != TILE_INVALID && p.isFOV())
	    {
		p.describeSquare(blind);
	    }

	    msg_newturn();
	}

	// Any other key stops the look.
	if (key)
	{
	    if (key == '\n' || key == ' ' || key == 'x')
	    {
		// Request to look closely
		lookclosely = true;
	    }
	    break;
	}
    }

    if (lookclosely)
    {
	// Check if we left on a mob.
	// Must re look up as displayWorld may have updated the map.
	p = glbDisplay->lookup(x, y);
	if (p.mob())
	{
	    displayMob(p.mob());
	}
	else if (p.item())
	{
	    displayItem(p.item());
	}
	else if (p.tile() != TILE_NONE)
	{
	    ATTR_NAMES	look = (ATTR_NAMES) p.defn().attr;

	    TILE_DEF	*tiledef = GAMEDEF::tiledef(p.tile());
	    // Make a new look if needed.
	    if (look < NUM_ATTRS)
	    {
		ATTR_NAMES	nlook = GAMEDEF::createNewAttrDef();
		ATTR_DEF	*nattr = GAMEDEF::attrdef(nlook);
		*nattr = *GAMEDEF::attrdef(look);
		tiledef->attr = nlook;
		look = nlook;
	    }
	    u8		symbol = p.defn().symbol;
	    BUF		name;
	    name.reference(p.defn().legend);

	    editLook("tile", name, symbol, look);

	    if (name.strcmp(tiledef->legend))
		tiledef->legend = glb_harden(name);
	    tiledef->symbol = symbol;
	}
    }

    glbInvertX = -1;
    glbInvertY = -1;
}

bool
doThrow(int itemno)
{
    int			dx, dy;

    msg_report("Throw in which direction?  ");
    if (awaitDirection(dx, dy) && (dx || dy))
    {
	msg_report(rand_dirtoname(dx, dy));
	msg_newturn();
    }
    else
    {
	msg_report("Cancelled.  ");
	msg_newturn();
	return false;
    }

    // Do the throw
    glbEngine->queue().append(COMMAND(ACTION_THROW, dx, dy, itemno));

    return true;
}

void
buildAction(ITEM *item, ACTION_NAMES *actions)
{
    glbChooser->clear();

    glbChooser->setTextAttr(ATTR_NORMAL);
    int			choice = 0;
    ACTION_NAMES	*origaction = actions;
    if (!item)
    {
	glbChooser->appendChoice("Nothing to do with nothing.");
	*actions++ = ACTION_NONE;
    }
    else
    {
	glbChooser->appendChoice("[ ] Do nothing");
	if (glbLastItemAction == ACTION_NONE)
	    choice = (int)(actions - origaction);
	*actions++ = ACTION_NONE;

	glbChooser->appendChoice("[X] Examine");
	if (glbLastItemAction == ACTION_EXAMINE)
	    choice = (int)(actions - origaction);
	*actions++ = ACTION_EXAMINE;

	if (item->isRanged())
	{
	    glbChooser->appendChoice("[B] Break");
	    if (glbLastItemAction == ACTION_BREAK)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_BREAK;
	}
	if (item->isPotion())
	{
	    glbChooser->appendChoice("[Q] Quaff");
	    if (glbLastItemAction == ACTION_QUAFF)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_QUAFF;
	    glbChooser->appendChoice("[T] Throw");
	    if (glbLastItemAction == ACTION_THROW)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_THROW;
	}
	if (item->isFood())
	{
	    glbChooser->appendChoice("[E] Eat");
	    if (glbLastItemAction == ACTION_EAT)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_EAT;
	}
	if (item->isMelee())
	{
	    glbChooser->appendChoice("[W] Wield");
	    if (glbLastItemAction == ACTION_WEAR)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_WEAR;
	}
	else if (item->isRing() || item->isArmour())
	{
	    glbChooser->appendChoice("[W] Wear");
	    if (glbLastItemAction == ACTION_WEAR)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_WEAR;
	}
	if (!item->defn().isflag)
	{
	    glbChooser->appendChoice("[D] Drop");
	    if (glbLastItemAction == ACTION_DROP)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_DROP;
	}
    }
    glbChooser->setChoice(choice);

    glbChooserActive = true;
}

bool
useItem(MOB *mob, ITEM *item, int itemno)
{
    ACTION_NAMES		actions[NUM_ACTIONS+1];
    bool			done = false;
    int				key;

    buildAction(item, actions);
    glbChooserActive = true;

    // Run the chooser...
    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();
	key = gfx_getKey(false);

	glbChooser->processKey(key);

	if (key)
	{
	    if (key == 'd' || key == 'D')
	    {
		glbEngine->queue().append(COMMAND(ACTION_DROP, itemno));
		done = true;
		break;
	    }
	    else if (key == 'q' || key == 'Q')
	    {
		glbEngine->queue().append(COMMAND(ACTION_QUAFF, itemno));
		done = true;
		break;
	    }
	    else if (key == 'e' || key == 'E')
	    {
		glbEngine->queue().append(COMMAND(ACTION_EAT, itemno));
		done = true;
		break;
	    }
	    else if (key == 'b' || key == 'B')
	    {
		glbEngine->queue().append(COMMAND(ACTION_BREAK, itemno));
		done = true;
		break;
	    }
	    else if (key == 'w' || key == 'W')
	    {
		glbEngine->queue().append(COMMAND(ACTION_WEAR, itemno));
		done = true;
		break;
	    }
	    else if (key == 'x' || key == 'X')
	    {
		displayItem(item);
		done = false;
		break;
	    }
	    else if (key == 't' || key == 'T')
	    {
		glbChooserActive = false;
		doThrow(itemno);
		done = true;
		break;
	    }
	    // User selects this?
	    else if (key == '\n' || key == ' ')
	    {
		ACTION_NAMES	action;

		action = actions[glbChooser->getChoice()];
		glbLastItemAction = action;
		switch (action)
		{
		    case ACTION_DROP:
			glbEngine->queue().append(COMMAND(ACTION_DROP, itemno));
			done = true;
			break;
		    case ACTION_QUAFF:
			glbEngine->queue().append(COMMAND(ACTION_QUAFF, itemno));
			done = true;
			break;
		    case ACTION_BREAK:
			glbEngine->queue().append(COMMAND(ACTION_BREAK, itemno));
			done = true;
			break;
		    case ACTION_WEAR:
			glbEngine->queue().append(COMMAND(ACTION_WEAR, itemno));
			done = true;
			break;
		    case ACTION_EAT:
			glbEngine->queue().append(COMMAND(ACTION_EAT, itemno));
			done = true;
			break;
		    case ACTION_SEARCH:
			glbEngine->queue().append(COMMAND(ACTION_SEARCH, itemno));
			done = true;
			break;
		    case ACTION_MEDITATE:
			glbEngine->queue().append(COMMAND(ACTION_MEDITATE, itemno));
			done = true;
			break;
		    case ACTION_EXAMINE:
			displayItem(item);
			done = false;
			break;
		    case ACTION_THROW:
			glbChooserActive = false;
			doThrow(itemno);
			done = true;
			break;
		}
		break;
	    }
	    else if (key == '\x1b')
	    {
		// Esc on options is to go back to play.
		// Play...
		break;
	    }
	    else
	    {
		// Ignore other options.
	    }
	}
    }
    glbChooserActive = false;

    return done;
}

void
buildInventory(MOB *mob)
{
    glbChooser->clear();

    glbChooser->setTextAttr(ATTR_NORMAL);
    if (!mob)
    {
	glbChooser->appendChoice("The dead own nothing.");
    }
    else
    {
	if (!mob->inventory().entries())
	{
	    glbChooser->appendChoice("nothing");
	}
	else
	{
	    for (int i = 0; i < mob->inventory().entries(); i++)
	    {
		ITEM 		*item = mob->inventory()(i);
		BUF		 name = item->getName();
		if (item->isEquipped())
		{
		    BUF		equip;
		    equip.sprintf("%s (equipped)", name.buffer());
		    name.strcpy(equip);
		}

		glbChooser->appendChoice(name);
	    }
	}
    }
    glbChooser->setChoice(glbLastItemIdx);

    glbChooserActive = true;
}

void
inventoryMenu()
{
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);
    int		key;
    bool	setlast = false;
    buildInventory(avatar);

    // Run the chooser...
    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();
	key = gfx_getKey(false);

	glbChooser->processKey(key);

	if (key)
	{
	    int		itemno = glbChooser->getChoice();
	    bool	done = false;
	    // User selects this?
	    if (key == '\n' || key == ' ')
	    {
		done = true;

		glbLastItemIdx = glbChooser->getChoice();
		setlast = true;
		if (itemno >= 0 && itemno < avatar->inventory().entries())
		{
		    done = useItem(avatar, avatar->inventory()(itemno), itemno);
		}
		// Finish the inventory menu.
		if (done)
		{
		    break;
		}
		else
		{
		    buildInventory(avatar);
		    setlast = false;
		}
	    }
	    else if (key == '\x1b' || key == 'i')
	    {
		// Esc on options is to go back to play.
		// Play...
		break;
	    }
	    else if (itemno >= 0 && itemno < avatar->inventory().entries())
	    {
		if (key == 'd' || key == 'D')
		{
		    glbEngine->queue().append(COMMAND(ACTION_DROP, itemno));
		    done = true;
		}
		else if (key == 'q' || key == 'Q')
		{
		    glbEngine->queue().append(COMMAND(ACTION_QUAFF, itemno));
		    done = true;
		}
		else if (key == 'e' || key == 'E')
		{
		    glbEngine->queue().append(COMMAND(ACTION_EAT, itemno));
		    done = true;
		}
		else if (key == 'b' || key == 'B')
		{
		    glbEngine->queue().append(COMMAND(ACTION_BREAK, itemno));
		    done = true;
		}
		else if (key == 'w' || key == 'W')
		{
		    glbEngine->queue().append(COMMAND(ACTION_WEAR, itemno));
		    done = true;
		}
		else if (key == 'x' || key == 'X')
		{
		    if (itemno >= 0 && itemno < avatar->inventory().entries())
		    {
			// This corrupts our chooser so we have
			// do abort.
			displayItem(avatar->inventory()(itemno));
			done = true;
		    }
		}
		else if (key == 't' || key == 'T')
		{
		    glbLastItemIdx = glbChooser->getChoice();
		    setlast = true;
		    glbChooserActive = false;
		    doThrow(itemno);
		    done = true;
		    break;
		}
	    }
	    else
	    {
		// Ignore other options.
	    }

	    if (done)
		break;
	}
    }
    if (!setlast)
	glbLastItemIdx = glbChooser->getChoice();
    glbChooserActive = false;
}

void
castSpell(MOB *mob, SPELL_NAMES spell)
{
    int			dx = 0;
    int			dy = 0;

    if (GAMEDEF::spelldef(spell)->needsdir)
    {
	msg_report("Cast in which direction?  ");
	if (awaitDirection(dx, dy))
	{
	    msg_report(rand_dirtoname(dx, dy));
	    msg_newturn();
	}
	else
	{
	    msg_report("Cancelled.  ");
	    msg_newturn();
	    return;
	}
    }
    glbLastSpell = spell;
    glbEngine->queue().append(COMMAND(ACTION_CAST, dx, dy, spell));
}

void
buildZapList(MOB *mob, PTRLIST<SPELL_NAMES> &list)
{
    glbChooser->clear();
    list.clear();
    int				chosen = 0;

    glbChooser->setTextAttr(ATTR_NORMAL);
    if (!mob)
    {
	glbChooser->appendChoice("The dead can cast nothing.");
	list.append(SPELL_NONE);
    }
    else
    {
	SPELL_NAMES		spell;

	for (spell = SPELL_NONE; spell < GAMEDEF::getNumSpell(); spell = (SPELL_NAMES) (spell + 1))
	{
	    ITEM_NAMES		itemname;
	    itemname = (ITEM_NAMES) GAMEDEF::spelldef(spell)->item;
	    if (itemname != ITEM_NONE &&
		mob->hasItem(itemname))
	    {
		BUF		buf;

		buf.sprintf("Cast %s", GAMEDEF::spelldef(spell)->name);
		glbChooser->appendChoice(buf);
		list.append(spell);
		if (spell == glbLastSpell)
		    chosen = list.entries()-1;
	    }
	}
    }
    if (!glbChooser->entries())
    {
	glbChooser->appendChoice("You know no magic spells.");
	list.append(SPELL_NONE);
    }

    glbChooser->setChoice(chosen);

    glbChooserActive = true;
}

void
zapMenu()
{
    PTRLIST<SPELL_NAMES>	spelllist;
    PTRLIST<int>		hotkeys;
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);

    buildZapList(avatar, spelllist);

    SPELL_NAMES	spell = runMenuSelection(spelllist, SPELL_NONE, hotkeys);

    if (spell != SPELL_NONE)
    {
	castSpell(avatar, spell);
    }
    glbChooserActive = false;
}

bool
selectWorld()
{
    int			key;
    bool		result = false;

    glbChooser->clear();
    glbChooser->setTextAttr(ATTR_NORMAL);

    // First save our game definition so it shows up in the 
    // list of options!
    {
	BUF		path;
	path.sprintf("../save/%s.txt", glbWorldName);
	GAMEDEF::save(path.buffer());
    }
    glbChooser->appendChoice("Create New World");

    PTRLIST<BUF>	worldlist;

    // Now list all the matching files.
    {
	DIRECTORY		dir;
	dir.opendir("../save");

	const char	*file;
	while (file = dir.readdir())
	{
	    BUF		world;
	    world.strcpy(file);

	    if (world.extensionMatch("txt"))
	    {
		// Defines a world...
		world.stripExtension();
		worldlist.append(world);
		glbChooser->appendChoice(world);
	    }
	}
	dir.closedir();
    }
    glbChooser->setChoice(0);
    glbChooserActive = true;

    // Run the chooser...
    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();
	key = gfx_getKey(false);

	glbChooser->processKey(key);

	if (key)
	{
	    // User selects this?
	    if (key == '\n' || key == ' ')
	    {
		// Save any current game, and the current world,
		// and generate a new one.
		const char		*newworld;

		if (glbChooser->getChoice() == 0)
		{
		    // New world
		    // Ask for the world name.
		    BUF		name;
		    name = popupQuestion("What do you wish to name your world?");
		    if (!name.isstring())
		    {
			// Cancel
			continue;
		    }
		    name.makeFileSafe();

		    newworld = strdup(name.buffer());
		}
		else
		{
		    // Load existing.
		    newworld = strdup(worldlist(glbChooser->getChoice()-1).buffer());
		}

		// Shutdown & restart.
		// Save the game.  A no-op if dead.
		glbEngine->queue().append(COMMAND(ACTION_SAVE));
		glbEngine->awaitSave();

		BUF		path;
		path.sprintf("../save/%s.txt", glbWorldName);
		GAMEDEF::save(path.buffer());

		// Clear out our map as it will become invalid with
		// the re-load.
		glbDisplay->clear();
		if (glbMap)
		    glbMap->decRef();
		glbMap = 0;

		glbWorldName = newworld;

		path.sprintf("../save/%s.txt", glbWorldName);
		GAMEDEF::load(path.buffer());

		// And now rebuild the level.
		glbEngine->queue().append(COMMAND(ACTION_RESTART, 1));
		// Update our flag so we show the correct welcome
		// message.
		glbReloadedLevel = glbEngine->awaitRebuild(false);

		glbChooserActive = false;
		result = true;
		break;
	    }
	    else if (key == '\x1b')
	    {
		// Esc on options is to go back to play.
		// Play...
		glbChooserActive = false;
		result = false;
		break;
	    }
	    else
	    {
		// Ignore other options.
	    }
	}
    }
    return result;
}

void
buildOptionsMenu(OPTION_NAMES d)
{
    OPTION_NAMES	option;
    glbChooser->clear();

    glbChooser->setTextAttr(ATTR_NORMAL);
    FOREACH_OPTION(option)
    {
	if (option == OPTION_PLAY)
	{
	    BUF		buf;
	    buf.sprintf("Play %s", glbWorldName);
	    glbChooser->appendChoice(buf);
	}
	else
	    glbChooser->appendChoice(glb_optiondefs[option].name);
    }
    glbChooser->setChoice(d);

    glbChooserActive = true;
}

bool
optionsMenu()
{
    int			key;
    bool		done = false;

    buildOptionsMenu(OPTION_PLAY);

    // Run the chooser...
    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();
	key = gfx_getKey(false);

	glbChooser->processKey(key);

	if (key)
	{
	    // User selects this?
	    if (key == '\n' || key == ' ')
	    {
		if (glbChooser->getChoice() == OPTION_INSTRUCTIONS)
		{
		    // Instructions.
		    popupText(text_lookup("game", "help"));
		}
		else if (glbChooser->getChoice() == OPTION_PLAY)
		{
		    // Play...
		    break;
		}
		else if (glbChooser->getChoice() == OPTION_CREATEWORLD)
		{
		    // Choose a new world to play in.
		    glbChooserActive = false;

		    selectWorld();

		    // Restore our options menu.
		    buildOptionsMenu(OPTION_CREATEWORLD);
		}
#if 0
		else if (glbChooser->getChoice() == OPTION_VOLUME)
		{
		    int			i;
		    BUF			buf;
		    // Volume
		    glbChooser->clear();

		    for (i = 0; i <= 10; i++)
		    {
			buf.sprintf("%3d%% Volume", (10 - i) * 10);

			glbChooser->appendChoice(buf);
		    }
		    glbChooser->setChoice(10 - glbConfig->myMusicVolume);

		    while (!TCODConsole::isWindowClosed())
		    {
			redrawWorld();
			key = gfx_getKey(false);

			glbChooser->processKey(key);

			setMusicVolume(10 - glbChooser->getChoice());
			if (key)
			{
			    break;
			}
		    }
		    buildOptionsMenu(OPTION_VOLUME);
		}
#endif
		else if (glbChooser->getChoice() == OPTION_COINS)
		{
		    glbConfig->myCoinsDynamic = !glbConfig->myCoinsDynamic;
		    glbCoinStack->setBarGraph(!glbConfig->myCoinsDynamic);
		}
		else if (glbChooser->getChoice() == OPTION_FULLSCREEN)
		{
		    glbConfig->myFullScreen = !glbConfig->myFullScreen;
		    // This is intentionally unrolled to work around a
		    // bool/int problem in libtcod
		    if (glbConfig->myFullScreen)
			TCODConsole::setFullscreen(true);
		    else
			TCODConsole::setFullscreen(false);
		}
		else if (glbChooser->getChoice() == OPTION_QUIT)
		{
		    // Quit
		    done = true;
		    break;
		}
	    }
	    else if (key == '\x1b')
	    {
		// Esc on options is to go back to play.
		// Play...
		break;
	    }
	    else
	    {
		// Ignore other options.
	    }
	}
    }
    glbChooserActive = false;

    return done;
}

bool
reloadLevel(bool alreadybuilt, bool alreadywaited = false, bool reload=false)
{
    if (TCODConsole::isWindowClosed())
	return true;

    glbVeryFirstRun = false;

    glbLevel = 1;
    msg_newturn();
    glbMessagePanel->clear();
    glbMessagePanel->appendText("> ");
    if (!alreadybuilt)
	glbEngine->queue().append(COMMAND(ACTION_RESTART, glbLevel));

    startMusic();

    // Wait for the map to finish.
    if (!alreadywaited)
	reload = glbEngine->awaitRebuild();
    if (reload)
    {
	welcomeText("Back");
    }
    else
    {
	welcomeText("You1");
    }

    // Redrawing the world updates our glbMap so everyone will see it.
    redrawWorld();

    return false;
}

// Let the avatar watch the end...
void
deathCoolDown()
{
    int		coolstartms = TCOD_sys_elapsed_milli();
    int		lastenginems = coolstartms;
    int		curtime;
    bool	done = false;
    double	ratio;

    while (!done)
    {
	curtime = TCOD_sys_elapsed_milli();

	// Queue wait events, one per 100 ms.
	while (lastenginems + 200 < curtime)
	{
	    glbEngine->queue().append(COMMAND(ACTION_WAIT));
	    lastenginems += 200;
	}

	// Set our fade.
	ratio = (curtime - coolstartms) / 2000.;
	ratio = 1.0 - ratio;
	if (ratio < 0)
	{
	    done = true;
	    ratio = 0;
	}

	TCOD_console_set_fade( (u8) (255 * ratio), TCOD_black);

	// Keep updating
	redrawWorld();
    }

    gfx_clearKeyBuf();

    TCOD_console_set_fade(255,  TCOD_black);
}

void
shutdownEverything()
{
    glbConfig->save("../7dayband.cfg");

    // Save the game.  A no-op if dead.
    glbEngine->queue().append(COMMAND(ACTION_SAVE));
    glbEngine->awaitSave();

    BUF		path;
    path.sprintf("../save/%s.txt", glbWorldName);
    GAMEDEF::save(path.buffer());

    stopMusic(true);
#ifdef USE_AUDIO
    if (glbTunes)
	Mix_FreeMusic(glbTunes);
#endif

    gfx_shutdown();

#ifdef USE_AUDIO
    SDL_QuitSubSystem(SDL_INIT_AUDIO);
#endif

#ifdef JOY_SUPPORT
    SDL_Quit();
#endif
}

void
queryOfferPrice(COMMAND &cmd)
{
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);
    if (!avatar || !avatar->alive())
    {
	cmd.action() = ACTION_NONE;
	return;
    }

    MOB		*at = avatar->map()->findMob(cmd.dy());
    ITEM	*item = avatar->getItemFromId(cmd.dz());
    if (!item || !at)
    {
	cmd.action() = ACTION_NONE;
	return;
    }

    BUF		query;
    BUF		itemname = item->getSingleName();

    query.sprintf("You notice %s has %d gold.\nHow much will you demand for %s?\n",
	    at->getName(),
	    at->getGold(),
	    itemname.buffer() );

    BUF		amount = popupQuestion(query);

    cmd.dw() = atoi(amount.buffer());
    if (cmd.dw() <= 0)
	cmd.action() = ACTION_NONE;
}

void
queryPurchasePrice(COMMAND &cmd)
{
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);
    if (!avatar || !avatar->alive())
    {
	cmd.action() = ACTION_NONE;
	return;
    }

    MOB		*at = avatar->map()->findMob(cmd.dy());
    ITEM	*item = at->getItemFromId(cmd.dz());

    if (!item || !at)
    {
	cmd.action() = ACTION_NONE;
	return;
    }
    BUF		query;
    BUF		itemname = item->getSingleName();

    query.sprintf("You have %d gold.\nHow much will you offer for %s?\n",
	    avatar->getGold(),
	    itemname.buffer() );

    BUF		amount = popupQuestion(query);

    cmd.dw() = atoi(amount.buffer());
    if (cmd.dw() <= 0)
	cmd.action() = ACTION_NONE;
}

void
buildShopMenu(MOB_NAMES keeper, PTRLIST<COMMAND> &list, PTRLIST<int> &hotkeys)
{
    glbChooser->clear();
    list.clear();
    BUF			buf;
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);
    bool	valid = false;

    if (!avatar || !avatar->alive())
	return;

    glbChooser->setPrequelAttr((ATTR_NAMES) glb_mobdefs[keeper].attr);

    if (valid)
    {
	glbChooser->appendChoice("No thank you.");
	list.append(COMMAND(ACTION_NONE));
	glbChooser->setChoice(list.entries()-1);
	glbChooserActive = true;
    }
    else
    {
	glbChooser->clear();
	list.clear();
    }
}


void
runShop(int shopkeeper_uid)
{
    PTRLIST<COMMAND>	shoppinglist;
    PTRLIST<int>	hotkeys;

    if (!glbMap)
	return;
    MOB		*shopkeeper_mob = glbMap->findMob(shopkeeper_uid);
    if (!shopkeeper_mob)
	return;

    // buildShopMenu(shopkeeper, shoppinglist, hotkeys);
    
    if (shoppinglist.entries() == 0)
	return;

    // Run the chooser...
    COMMAND	cmd = runMenuSelection(shoppinglist, COMMAND(ACTION_NONE), hotkeys);

    if (cmd.action() != ACTION_NONE)
    {
	glbEngine->queue().append(cmd);
    }
}

#ifdef WIN32
int WINAPI
WinMain(HINSTANCE hINstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCMdSHow)
#else
int 
main(int argc, char **argv)
#endif
{
    bool		done = false;

    // Clamp frame rate.
    /// Has to be high to fix keyrepeat problem :<
    TCOD_sys_set_fps(200);

    glbConfig = new CONFIG();
    glbConfig->load("../7dayband.cfg");

    // Dear Microsoft,
    // The following code in optimized is both a waste and seems designed
    // to ensure that if we call a bool where the callee pretends it is
    // an int we'll end up with garbage data.
    //
    // 0040BF4C  movzx       eax,byte ptr [eax+10h] 

    // TCODConsole::initRoot(SCR_WIDTH, SCR_HEIGHT, "The Smith's Hand", myFullScreen);
    // 0040BF50  xor         ebp,ebp 
    // 0040BF52  cmp         eax,ebp 
    // 0040BF54  setne       cl   
    // 0040BF57  mov         dword ptr [esp+28h],eax 
    // 0040BF5B  push        ecx  

    TCODConsole::setCustomFont("terminal.png", TCOD_FONT_LAYOUT_ASCII_INCOL | TCOD_FONT_TYPE_GREYSCALE);

    // My work around is to constantify the fullscreen and hope that
    // the compiler doesn't catch on.
    if (glbConfig->myFullScreen)
	TCODConsole::initRoot(SCR_WIDTH, SCR_HEIGHT, "Seven Day Band", true);
    else
	TCODConsole::initRoot(SCR_WIDTH, SCR_HEIGHT, "Seven Day Band", false);

    // TCOD doesn't do audio.
#ifdef USE_AUDIO
    SDL_InitSubSystem(SDL_INIT_AUDIO);
    if (Mix_OpenAudio(22050, AUDIO_S16, 2, 4096))
    {
	printf("Failed to open audio!\n");
	exit(1);
    }
#endif

#ifdef JOY_SUPPORT
    SDL_InitSubSystem(SDL_INIT_JOYSTICK);
#endif

    setMusicVolume(glbConfig->musicVolume());

    rand_setseed((long) time(0));

    {
	BUF		path;
	path.sprintf("../save/%s.txt", glbWorldName);
	GAMEDEF::load(path.buffer());
    }

    gfx_init();
    MAP::init();

    text_init();

    glbDisplay = new DISPLAY(0, 0, SCR_WIDTH, SCR_HEIGHT-5, 80);

    glbMessagePanel = new PANEL(40, 5);
    msg_registerpanel(glbMessagePanel);
    glbMessagePanel->move(20, SCR_HEIGHT-5);

    glbPopUp = new PANEL(50, 20);
    glbPopUp->move(15, 5);
    glbPopUp->setBorder(true, ' ', ATTR_BORDER);
    // And so the mispelling is propagated...
    glbPopUp->setRigthMargin(1);
    glbPopUp->setIndent(1);

    glbJacob = new PANEL(30, 25);
    glbJacob->move(55, 0);
    glbJacob->setTransparent(true);

    glbWeaponInfo = new PANEL(30, 30);
    glbWeaponInfo->move(1, 2);
    glbWeaponInfo->setTransparent(true);

    glbVictoryInfo = new PANEL(26, 5);
    glbVictoryInfo->move(28, 5);
    glbVictoryInfo->setIndent(2);
    glbVictoryInfo->setBorder(true, ' ', ATTR_VICTORYBORDER);
    glbVictoryInfo->clear();
    glbVictoryInfo->newLine();
    glbVictoryInfo->appendText("The orcs are all dead!");
    glbVictoryInfo->newLine();
    glbVictoryInfo->newLine();
    glbVictoryInfo->appendText("Press [V] to finish.");

    glbLevel = 1;

    glbChooser = new CHOOSER();
    glbChooser->move(SCR_WIDTH/2, SCR_HEIGHT/2, CHOOSER::JUSTCENTER, CHOOSER::JUSTCENTER);
    glbChooser->setBorder(true, ' ', ATTR_BORDER);
    glbChooser->setIndent(1);
    glbChooser->setRigthMargin(1);
    glbChooser->setAttr(ATTR_NORMAL, ATTR_HILITE, ATTR_LIGHTBLACK, ATTR_DKHILITE);

    glbLevelBuildStatus = new CHOOSER();
    glbLevelBuildStatus->move(SCR_WIDTH/2, SCR_HEIGHT/2, CHOOSER::JUSTCENTER, CHOOSER::JUSTCENTER);
    glbLevelBuildStatus->setBorder(true, ' ', ATTR_BORDER);
    glbLevelBuildStatus->setIndent(1);
    glbLevelBuildStatus->setRigthMargin(1);
    glbLevelBuildStatus->setAttr(ATTR_NORMAL, ATTR_HILITE, ATTR_HILITE,
				ATTR_HILITE);

    glbEngine = new ENGINE(glbDisplay);

    glbHealthFire = new FIREFLY(0, 1, SCR_HEIGHT, FIRE_HEALTH, false);
    glbHealthFire->setBarGraph(true);
    glbManaFire = new FIREFLY(0, 1, SCR_HEIGHT, FIRE_ICE, false);
    glbManaFire->setBarGraph(true);
    glbFoodFire = new FIREFLY(0, 1, SCR_HEIGHT, FIRE_FOOD, false);
    glbFoodFire->setBarGraph(true);
    glbCoinStack = new FIREFLY(0, 20, SCR_HEIGHT, FIRE_COLOR, false);
    glbCoinStack->setBarGraph(!glbConfig->myCoinsDynamic);

    {
	ATTR_DEF	*def = GAMEDEF::attrdef(GAMEDEF::itemdef(ITEM_GOLD)->attr);
	glbCoinStack->setColor( def->fg_r, def->fg_g, def->fg_b );
    }

    // Run our first level immediately so it can be built or 
    // loaded from disk while people wait.
    // The problem is we expect to be able to create/restore
    // worlds from the options menu, so we had better have
    // finished it prior to that!
    glbEngine->queue().append(COMMAND(ACTION_RESTART, 1));

    glbReloadedLevel = glbEngine->awaitRebuild();

    done = optionsMenu();
    if (done)
    {
	shutdownEverything();
	return 0;
    }

    done = reloadLevel(true, true, glbReloadedLevel);
    if (done)
    {
	shutdownEverything();
	return 0;
    }

    do
    {
	int		key;
	int		dx, dy;

	redrawWorld();

	while (1)
	{
	    BUF		pop;

	    pop = glbEngine->getNextPopup();
	    if (!pop.isstring())
		break;

	    popupText(pop, POPUP_DELAY);
	}

	while (1)
	{
	    ENGINE::Question		query;

	    if (glbEngine->getQuestion(query))
	    {
		glbEngine->answerQuestion(defedit_askQuestion(query));
	    }
	    else
		break;
	}


	if (glbMap && glbMap->avatar() &&
	    !glbMap->avatar()->alive())
	{
	    // You have died...
	    deathCoolDown();

	    BUF			 tmp;
	    ITEM		*gold = glbMap->avatar()->lookupItem(ITEM_GOLD);
	    if (!gold)
	    {
		tmp.sprintf("lack of %s", GAMEDEF::itemdef(ITEM_GOLD)->name);
		text_registervariable("GOLD", tmp);
	    }
	    else
	    {
		text_registervariable("GOLD", gold->getName());
	    }
	    text_registervariable("ITEM", GAMEDEF::itemdef(ITEM_MACGUFFIN)->name);
	    text_registervariable("MOBNAME", GAMEDEF::mobdef(MOB_BOSS)->name);

	    popupText(text_lookup("game", "lose"));

	    stopMusic();

	    done = optionsMenu();
	    if (done)
	    {
		shutdownEverything();
		return 0;
	    }
	    // Don't prep a reload.
	    done = reloadLevel(false);
	}
	else if (glbMap && glbMap->avatar() &&
		glbMap->avatar()->hasWon())
	{
	    // Traditional win.
	    BUF			 tmp;
	    ITEM		*gold = glbMap->avatar()->lookupItem(ITEM_GOLD);
	    if (!gold)
	    {
		tmp.sprintf("lack of %s", GAMEDEF::itemdef(ITEM_GOLD)->name);
		text_registervariable("GOLD", tmp);
	    }
	    else
	    {
		text_registervariable("GOLD", gold->getName());
	    }
	    text_registervariable("ITEM", GAMEDEF::itemdef(ITEM_MACGUFFIN)->name);
	    text_registervariable("MOBNAME", GAMEDEF::mobdef(MOB_BOSS)->name);
	    popupText(text_lookup("game", "victory"));

	    stopMusic();
	    done = optionsMenu();
	    if (done)
	    {
		shutdownEverything();
		return 0;
	    }
	    done = reloadLevel(false);
	}
	else if (glbMap && !glbMap->avatar())
	{
	    stopMusic();
	    done = optionsMenu();
	    if (done)
	    {
		shutdownEverything();
		return 0;
	    }
	    done = reloadLevel(false);
	}

	key = gfx_getKey(false);

	if (gfx_cookDir(key, dx, dy))
	{
	    // Direction.
	    glbEngine->queue().append(COMMAND(ACTION_BUMP, dx, dy));
	}

	switch (key)
	{
	    case 'Q':
		done = true;
		break;

	    case 'R':
		done = reloadLevel(false);
		break;

	    case 'w':
		glbEngine->queue().append(COMMAND(ACTION_WAITUNTIL));
		break;

	    case '>':
	    case '<':
		glbEngine->queue().append(COMMAND(ACTION_CLIMB));
		break;

	    case 'i':
		inventoryMenu();
		break;

	    case 'a':
		popupText(text_lookup("game", "about"));
		break;

	    case 'W':
	    {
		welcomeText("You1");
		break;
	    }

	    case 'P':
		glbConfig->myFullScreen = !glbConfig->myFullScreen;
		// This is intentionally unrolled to work around a
		// bool/int problem in libtcod
		if (glbConfig->myFullScreen)
		    TCODConsole::setFullscreen(true);
		else
		    TCODConsole::setFullscreen(false);
		break;
	
	    case GFX_KEYF1:
	    case '?':
		popupText(text_lookup("game", "help"));
		break;

	    case 'x':
		doExamine();
		break;

/*
	    case 'f':
	    case '0':
		msg_report("Shoot in what direction?  ");
		if (awaitDirection(dx, dy) && (dx || dy))
		{
		    msg_report(rand_dirtoname(dx, dy));
		    msg_newturn();
		    glbEngine->queue().append(COMMAND(ACTION_FIRE, dx, dy));
		}
		else
		{
		    msg_report("Cancelled.  ");
		    msg_newturn();
		}
		break;
*/

	    case 'z':
	    case '+':
	    {
		zapMenu();
		break;
	    }

	    case 's':
		glbEngine->queue().append(COMMAND(ACTION_SEARCH));
		break;


#if 0
	    case 'W':
		glbEngine->queue().append(COMMAND(ACTION_CREATEITEM, ITEM_BOW));

		break;
#endif

	    case '\x1b':
		// If meditating, abort.
		if (glbMap && glbMap->avatar() && 
		    glbMap->avatar()->alive() && glbMap->avatar()->isMeditating())
		{
		    glbEngine->queue().append(COMMAND(ACTION_MEDITATE));
		    break;
		}
		// If waiting, abort.
		if (glbMap && glbMap->avatar() && 
		    glbMap->avatar()->alive() && glbMap->avatar()->isWaiting())
		{
		    glbEngine->queue().append(COMMAND(ACTION_STOPWAITING));
		    break;
		}
		// FALL THROUGH
	    case 'O':
		done = optionsMenu();
		break;
	}
    } while (!done && !TCODConsole::isWindowClosed());

    shutdownEverything();

    return 0;
}
